import hashlib


class HashState:
    """Utility class for hashing data. Uses ``hashlib.md5``.

    :example usage:
    .. code-block:: python

        >>> state = HashState()
        >>> state.update(b"hello world")
        >>> state.state()
        'fe0cf2fe0d7cb366190a4a80af973909'
    """

    def __init__(self):
        self._state = hashlib.md5()

    def update(self, data: bytes):
        """Append some data to the current state.

        :param data: the data to append
        :type data: bytes

        :return: self
        """
        self._state.update(str(len(data)).encode("utf-8") + b":")
        self._state.update(data + b":")
        return self

    def record(self, *args):
        """Append multiple pieces of data to the current state.

        :param args: the data to append
        :type args: *str

        :return: self
        """
        self._state.update(str(args).encode("utf-8"))
        return self

    def state(self):
        """Get a hex digest of the current state.

        :return: the hex digest of the current state
        """
        return self._state.hexdigest()



from functools import wraps

from flask import Flask


def job(app: Flask, endpoint):
    """Adds a URL rule for a recurring job at ``/jobs/<endpoint>``.

    :param app: the app the method belongs to
    :type app: ~flask.Flask

    :param endpoint: the endpoint to route to the method
    :type endpoint: str

    :return: a decorator which can be applied to a function to bind it to
        ``/jobs/<endpoint>``
    """

    def decorator(func):
        @wraps(func)
        def wrapped():
            func()
            return ""

        app.add_url_rule(
            f"/jobs/{endpoint}",
            func.__name__ + "_job",
            wrapped,
            strict_slashes=False,
            methods=["POST"],
        )

        return func

    return decorator



from colorama import Style


def pretty_print(emoji: str, msg: str):
    """Pretty prints a message.

    :param emoji: the emoji to wrap the message in
    :type emoji: str
    :param msg: the message to wrap with the emoji
    :type msg: str
    """
    print(f"{emoji}{Style.BRIGHT} {msg} {Style.RESET_ALL}{emoji}")



import os
import sys
from contextlib import contextmanager
from os import getenv
from os.path import exists
from time import sleep
from typing import List

import sqlalchemy.engine.url

from common.rpc.secrets import get_secret

use_devdb = getenv("ENV", "DEV") in ("DEV", "TEST")
use_prod_proxy = getenv("ENV") == "DEV_ON_PROD"

is_sphinx = "sphinx" in sys.argv[0]

if use_devdb:
    database_url = "sqlite:///" + os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../app.db"
    )
elif use_prod_proxy:
    curr = os.path.abspath(os.path.curdir)
    app = os.path.basename(curr)
    prev = None
    while True:
        if ".git" in os.listdir(curr):
            break
        prev = curr
        curr = os.path.dirname(curr)
    assert (
        input(
            "Are you sure? You are about to use the PROD database on a DEV build (y/N) "
        )
        .strip()
        .lower()
        == "y"
    )
    if "cloud_sql_proxy" not in os.listdir(curr):
        print(
            "Follow the instructions at https://cloud.google.com/sql/docs/mysql/sql-proxy to install the proxy "
            "in the root directory of the repository, then try again."
        )
        exit(0)
    database_url = sqlalchemy.engine.url.URL(
        drivername="mysql+pymysql",
        username="buildserver",
        password=get_secret(secret_name="DATABASE_PW"),
        host="127.0.0.1",
        port=3307,
        database=app.replace("-", "_"),
    ).__to_string__(hide_password=False)
else:
    database_url = getenv("DATABASE_URL")

engine = sqlalchemy.create_engine(
    database_url,
    **(
        dict(
            connect_args=dict(
                ssl=dict(
                    cert="/shared/client-cert.pem",
                    key="/shared/client-key.pem",
                    ca="/shared/server-ca.pem",
                ),
            )
        )
        if exists("/shared/client-cert.pem")
        else {}
    ),
    **(
        {}
        if use_devdb
        else dict(pool_size=5, max_overflow=2, pool_timeout=30, pool_recycle=1800)
    ),
)


@contextmanager
def connect_db(*, retries=3):
    """Create a context that uses a connection to the current database.

    :param retries: the number of times to try connecting to the database
    :type retries: int

    :yields: a function with parameters ``(query: str, args: List[str] = [])``,
        where the ``query_str`` should use ``%s`` to represent sequential
        arguments in the ``args_list``

    :example usage:
    .. code-block:: python

        with connect_db() as db:
            db("INSERT INTO animals VALUES %s", ["cat"])
            output = db("SELECT * FROM animals")
    """
    if is_sphinx:

        def no_op(*args, **kwargs):
            class NoOp:
                fetchone = lambda *args: None
                fetchall = lambda *args: None

            return NoOp()

        yield no_op
    else:
        for i in range(retries):
            try:
                conn = engine.connect()
                break
            except:
                sleep(3)
                continue
        else:
            raise

        with conn:

            def db(query: str, args: List[str] = []):
                if use_devdb:
                    query = query.replace("%s", "?")
                return conn.execute(query, args)

            yield db


@contextmanager
def transaction_db():
    """Create a context that performs a transaction on current database.

    The difference between this and :meth:`~common.db.connect_db` is that this
    method batches queries in a transaction, and only runs them once the
    context is abandoned.

    :yields: a function with parameters ``(query: str, args: List[str] = [])``,
        where the ``query_str`` should use ``%s`` to represent sequential
        arguments in the ``args_list``

    :example usage:
    .. code-block:: python

        with transaction_db() as db:
            db("INSERT INTO animals VALUES %s", ["cat"])
            output = db("SELECT * FROM animals")
    """
    with engine.begin() as conn:

        def db(*args):
            query, *rest = args
            if use_devdb:
                query = query.replace("%s", "?")
            return conn.execute(query, *rest)

        yield db



from typing import Union

from common.rpc.utils import (
    create_service,
    requires_access_token,
    requires_master_secret,
)

service = create_service(__name__)


@requires_master_secret
@service.route("/api/list_channels")
def list_channels(*, course: str):
    ...


@requires_master_secret
@service.route("/api/post_message")
def post_message(*, course: str, message: Union[str, dict], channel: str):
    ...


@requires_access_token
@service.route("/api/list_conversations")
def list_conversations(*, course: str, include_dms: bool = False):
    ...


@requires_access_token
@service.route("/api/export_channel")
def export_channel(*, course: str, channel_id: str):
    ...



from typing import Tuple

from common.rpc.utils import create_service, requires_master_secret

service = create_service("buildserver-hosted-worker")


@requires_master_secret
@service.route("/api/build_worker_build")
def build_worker_build(*, pr_number: int, sha: str) -> Tuple[bool, str]:
    ...


@requires_master_secret
@service.route("/api/build_worker_run_checks")
def build_worker_run_checks(*, sha: str) -> Tuple[bool, str]:
    ...



import json
import os
import traceback
from functools import wraps
from sys import stderr
from time import sleep
from typing import List
from urllib.error import HTTPError

import flask
import requests
from cachetools import TTLCache
from flask import Response, has_request_context, jsonify, request, stream_with_context

from common.rpc.auth_utils import get_token, refresh_token
from common.secrets import get_master_secret

STATUS_MARKER = "__INTERNAL_STATUS_MARKER"
GCP_INTERNAL_ERROR_CODE = 503


class Service:
    def __init__(self, route):
        self.route = route


def find_default_endpoints(app: str, path: str):
    endpoints = []
    if has_request_context():
        proxied_host = request.headers.get("X-Forwarded-For-Host")
        if proxied_host:
            parts = proxied_host.split(".")
            if "pr" in parts:
                pr = parts[0]
                endpoints.append(f"https://{pr}.{app}.pr.cs61a.org{path}")
    endpoints.append(f"https://{app}.cs61a.org{path}")
    return endpoints


def select_endpoint(endpoints: List[str], path: str, retries: int):
    # try all the PR candidates
    for i, endpoint in enumerate(endpoints[:-1]):
        try:
            for _ in range(retries + 1):
                # check if the PR / endpoint exists
                check_exists = requests.get(
                    endpoint[: -len(path)],
                    allow_redirects=False,
                )
                if check_exists.status_code == GCP_INTERNAL_ERROR_CODE:
                    # this error is not our fault, retry after a short pause
                    sleep(1)
                    continue
                check_exists.raise_for_status()
                return endpoint
            else:
                # if we exhaust all our retries, give up on this endpoint
                continue
        except (HTTPError, requests.ConnectionError):
            continue

    # fall back to prod
    return endpoints[-1]


def stream_encode(out):
    try:
        for x in out:
            yield bytes(x, encoding="ascii", errors="replace")
    except Exception as e:
        yield STATUS_MARKER
        yield bytes(str(e), encoding="ascii", errors="replace")
    else:
        yield STATUS_MARKER


def receive_stream(resp: Response):
    buffer = []
    ok = True
    for x in resp.iter_content():
        buffer.append(x.decode("ascii"))
        buff_string = "".join(buffer)
        if STATUS_MARKER in buff_string:
            # We are now reading in the error message
            # Stop flushing the buffer
            ok = False
        if ok and len(buff_string) > len(STATUS_MARKER):
            yield buff_string[: -len(STATUS_MARKER)]
            buffer = [buff_string[-len(STATUS_MARKER) :]]
    buff_string = "".join(buffer)
    if not buff_string.endswith(STATUS_MARKER):
        # some error occurred
        pos = buff_string.index(STATUS_MARKER)
        raise Exception(buff_string[pos + len(STATUS_MARKER) :])
    yield from buffer[: -len(STATUS_MARKER)]


def create_service(app: str, override=None, providers=None):
    app = override or app.split(".")[-1]

    def route(path, *, streaming=False):
        def decorator(func):
            @wraps(func)
            def wrapped(*, noreply=False, timeout=1, retries=0, local=False, **kwargs):
                assert (
                    not noreply or not retries
                ), "Cannot retry a noreply request, use streaming instead"

                if local:
                    return wrapped.local(**kwargs)

                if providers:
                    endpoints = [f"{provider}{path}" for provider in providers]
                else:
                    endpoints = find_default_endpoints(app, path)

                endpoint = select_endpoint(endpoints, path, retries)

                if noreply:
                    try:
                        requests.post(endpoint, json=kwargs, timeout=timeout)
                    except requests.exceptions.ReadTimeout:
                        return
                else:
                    for _ in range(retries + 1):
                        resp = requests.post(endpoint, json=kwargs, stream=streaming)
                        if resp.status_code == GCP_INTERNAL_ERROR_CODE:
                            sleep(1)
                            continue
                        break
                    else:
                        # we exhausted all our retries
                        resp.raise_for_status()

                    if resp.status_code == 401:
                        raise PermissionError(resp.text)
                    elif resp.status_code == 500:
                        raise Exception(resp.text)
                    resp.raise_for_status()
                    if streaming:
                        return receive_stream(resp)
                    else:
                        return resp.json()

            def bind(app: flask.Flask):
                def decorator(func):
                    def handler():
                        kwargs = request.json
                        try:
                            out = func(**kwargs)
                            if streaming:
                                return Response(stream_with_context(stream_encode(out)))
                            else:
                                return jsonify(out)
                        except PermissionError as e:
                            return "", 401
                        except Exception as e:
                            traceback.print_exc(file=stderr)
                            print(str(e), file=stderr)
                            return "", 500

                    app.add_url_rule(path, func.__name__, handler, methods=["POST"])
                    wrapped.local = func
                    return func

                return decorator

            wrapped.bind = bind

            return wrapped

        return decorator

    return Service(route)


def requires_master_secret(func):
    @wraps(func)
    def wrapped(*, _impersonate=None, _sudo_token=None, **kwargs):
        if _sudo_token:
            return func(**kwargs, _impersonate=_impersonate, _sudo_token=_sudo_token)
        elif not get_master_secret() and _impersonate and not _sudo_token:
            from common.rpc.secrets import (
                get_secret_from_server,
            )  # placed here to avoid circular imports

            print(f"Attempting to impersonate {_impersonate}")

            try:
                sudo_secret = get_secret_from_server(
                    secret_name="MASTER",
                    _impersonate=_impersonate,
                    _sudo_token=get_token(),
                )
            except PermissionError:
                refresh_token()
                try:  # second attempt, in case the first was just an expired token
                    sudo_secret = get_secret_from_server(
                        secret_name="MASTER",
                        _impersonate=_impersonate,
                        _sudo_token=get_token(),
                    )
                except PermissionError:
                    raise PermissionError(
                        "You must be logged in as an admin to do that."
                    )

            master_secret = sudo_secret
        else:
            master_secret = get_master_secret()

        return func(**kwargs, master_secret=master_secret)

    return wrapped


def requires_access_token(func):
    @wraps(func)
    def wrapped(**kwargs):
        if os.getenv("ENV") == "prod":
            return func(**kwargs, master_secret=get_master_secret())
        try:
            return func(**kwargs, access_token=get_token())
        except PermissionError:
            refresh_token()
            return func(**kwargs, access_token=get_token())

    return wrapped


def cached(ttl: int = 1800):
    """
    Caches the return value of this RPC method for `ttl` seconds (defaults to 1800s)
    """
    cache = TTLCache(1000, ttl)

    def decorator(func):
        @wraps(func)
        def wrapped(**kwargs):
            key = json.dumps(kwargs)
            if key not in cache:
                cache[key] = func(**kwargs)
            return cache[key]

        return wrapped

    return decorator



from typing import List, Dict, Optional
from common.rpc.utils import (
    create_service,
    requires_master_secret,
    requires_access_token,
)

service = create_service(__name__, "ag-master")


@requires_master_secret
@service.route("/api/prepare_jobs")
def prepare_jobs(job_files: List[Dict[str, str]]) -> List[str]:
    """
    Takes a list of file name -> file content mappings, and schedules jobs for each mapping in the list.
    Returns a list of the scheduled job IDs, to be passed to trigger_jobs
    """
    ...


@requires_master_secret
@service.route("/api/trigger_jobs")
def trigger_jobs(command, batch_size, skeleton_path, jobs, script_timeout):
    """Given a list of job IDs, queues autograder workers to begin grading.

    :param command: the command that runs each job
    :type command: str
    :param batch_size: the number of jobs to be run together in a single batch
    :type batch_size: int
    :param skeleton_path: the path to the cloud storage blob holding the skeleton files for this job
    :type skeleton_path: str
    :param jobs: the list of job IDs to queue
    :type jobs: list[str]
    :param script_timeout: the maximum running time of each job, in seconds
    :type script_timeout: optional[int]

    Bound in :func:`~docs.ag_master.okpy.create_okpy_endpoints`
    """
    ...


@service.route("/api/get_results")
def get_results(job_ids: List[str]) -> Dict[str, Optional[Dict[str, str]]]:
    ...


@service.route("/api/get_submission")
def get_submission(*, job_id: str) -> Dict:
    """Given a job ID, gets submission information from Okpy and forwards it
    to the requesting worker.

    :param job_id: the job ID associated with the requested backup
    :type job_id: str

    :return: a dictionary containing backup information, specifically the
        ``data`` value `here <https://okpy.github.io/documentation/ok-api.html#backups-view-a-backup>`_

    Bound in :func:`~docs.ag_master.worker.create_worker_endpoints`
    """
    ...


@service.route("/api/handle_output")
def handle_output(*, output: str, job_id: str):
    """Given output and a job ID, parses the output for scores and uploads them
    to Okpy. See :class:`~docs.ag_master.worker` for helper functions used.

    :param output: the autograder worker output
    :type output: str
    :param job_id: the job ID associated with the output
    :type job_id: str

    Bound in :func:`~docs.ag_master.worker.create_worker_endpoints`
    """
    ...


@service.route("/api/set_failure")
def set_failure(*, job_id: str, result: str):
    """Given a job ID and some output result, set the job to failed and report
    the failure to Okpy.

    :param job_id: the job ID associated with the output
    :type job_id: str
    :param result: the autograder worker output
    :type result: str

    Bound in :func:`~docs.ag_master.worker.create_worker_endpoints`
    """
    ...


@requires_access_token
@service.route("/api/upload_zip")
def upload_zip(*, course: str, name: str, file: str):
    """Given a course, a filename, and base64-encoded zip file contents,
    upload the file to a cloud storage bucket. This method is meant to be
    accessed using the SICP command-line tool.

    :param course: the course code
    :type course: str
    :param name: the filename
    :type name: str
    :param file: base64-encoded zip-file contexts
    :type file: str

    Bound in :func:`~docs.ag_master.admin.create_admin_endpoints`
    """
    ...


@requires_access_token
@service.route("/api/create_assignment")
def create_assignment(
    *,
    course: str,
    name: str,
    file: str,
    command: str,
    batch_size: int,
    grading_base: str,
) -> str:
    """Given some information about an assignment, register it into the
    autograder database so that it may be graded using the autograder. This
    method is meant to be accessed using the SICP command-line tool.

    :param course: the course code
    :type course: str
    :param name: the assignment shortname
    :type name: str
    :param file: the assignment filename (should be the same as ``name`` in
        :func:`~common.rpc.ag_master.upload_zip`)
    :type file: str
    :param command: the command the worker should run to grade a backup for this
        assignment
    :type command: str
    :param batch_size: how many backups should be graded by one worker
    :type batch_size: int
    :param grading_base: the grading server associated with the assignment
        (by default, https://okpy.org)
    :type grading_base: str

    Bound in :func:`~docs.ag_master.admin.create_admin_endpoints`
    """
    ...



from os import getenv

from common.rpc.auth import can_user_rpc
from common.course_config import get_course
from common.oauth_client import get_user, is_staff


def can_user(action: str):
    if getenv("ENV") != "prod":
        return True

    if get_course() == "cs61a":
        return can_user_rpc(
            course=get_course(),
            email=get_user()["email"],
            action=action,
        )
    else:
        return is_staff(get_course())



import re, requests
from os import getenv

from cachetools import TTLCache
from flask import request
from common.rpc import auth

DOMAIN_COURSES = TTLCache(1000, 1800)
COURSE_ENDPOINTS = TTLCache(1000, 1800)
ENDPOINT_ID = TTLCache(1000, 1800)


def get_course(domain=None):
    """Gets the course code of the course that owns the current domain, using
    :meth:`~common.rpc.auth.get_domain`.

    :param domain: the domain name in question, inferred using
        :meth:`~common.course_config.get_domain` if omitted
    :type domain: str

    :return: the course code, such as "cs61a"
    """
    if getenv("ENV") != "prod":
        return "cs61a"
    if not domain:
        domain = get_domain()
    if "pr" in domain:
        DOMAIN_COURSES[domain] = "cs61a"
    if domain not in DOMAIN_COURSES:
        DOMAIN_COURSES[domain] = auth.get_course(domain=domain)
    return DOMAIN_COURSES[domain]


def get_domain():
    """Gets the domain that this request is being made on.

    :return: the domain name, based on request headers
    """
    return request.headers.get("X-Forwarded-For-Host") or request.headers["HOST"]


def get_endpoint(course=None):
    """Gets a course's most recent Okpy endpoint, typically the current
    semester's, using :meth:`~common.rpc.auth.get_endpoint`.

    :param course: the course code, such as "cs61a", inferred using
        :meth:`~common.course_config.get_course` if omitted
    :type course: str

    :return: the Okpy endpoint, such as "cal/cs61a/sp21"
    """
    if getenv("ENV") != "prod":
        return "cal/cs61a/fa21"
    if not course:
        course = get_course()
    if course not in COURSE_ENDPOINTS:
        COURSE_ENDPOINTS[course] = auth.get_endpoint(course=course)
    return COURSE_ENDPOINTS[course]


def get_course_id(course=None):
    """Gets a course's most recent Okpy course ID, typically the current
    semester's, using :meth:`~common.rpc.auth.get_endpoint_id`.

    :param course: the course code, such as "cs61a", inferred using
        :meth:`~common.course_config.get_course` if omitted
    :type course: str

    :return: the Okpy endpoint, such as 707
    """
    if getenv("ENV") != "prod":
        return 1
    if not course:
        course = get_course()
    if course not in ENDPOINT_ID:
        ENDPOINT_ID[course] = auth.get_endpoint_id(course=course)
    return ENDPOINT_ID[course]


def is_admin(email, course=None):
    """Returns whether or not an email address belongs to an admin
    for the given course, using :meth:`~common.rpc.auth.is_admin`.

    :param email: the email address in question
    :type email: str
    :param course: the course code, such as "cs61a", inferred using
        :meth:`~common.course_config.get_course` if omitted
    :type course: str

    :return: ``True`` if the user is an admin, ``False`` otherwise
    """
    if getenv("ENV") != "prod":
        return True
    if not course:
        course = get_course()
    return auth.is_admin(email=email, course=course, force_course=course)


def get_user_from_token(access_token):
    ret = requests.get(
        "https://okpy.org/api/v3/user/", params={"access_token": access_token}
    )
    if ret.status_code != 200:
        raise PermissionError
    return ret.json()["data"]


def is_admin_token(access_token, course=None):
    """Returns whether or not an Okpy access token belongs to an admin
    for the given course.

    :param access_token: the Okpy access token in question
    :type access_token: str
    :param course: the course code, such as "cs61a", inferred using
        :meth:`~common.course_config.get_course` if omitted
    :type course: str

    :return: ``True`` if the user is an admin, ``False`` otherwise
    """
    return is_admin(
        get_user_from_token(access_token)["email"],
        course=course,
    )


def format_coursecode(course):
    """Formats a course code is a pretty way, separating the department from
    the course number.

    :param course: the course code, such as "cs61a"
    :type course: str

    :return: prettified course code, such as "CS 61A"
    """
    m = re.match(r"([a-z]+)([0-9]+[a-z]?)", course)
    return m and (m.group(1) + " " + m.group(2)).upper()



from json import dumps, loads
from sys import stderr

import click

from common.rpc.slack import export_channel, list_conversations


@click.command()
@click.option("--course", prompt=True, required=True)
@click.option("--include-dms/--exclude-dms", prompt=True, required=True)
@click.option("--load-from", type=click.File("r"))
@click.option("--out", type=click.File("w"), prompt=True, required=True)
def slack_export(course, include_dms, load_from, out):
    """Export all the data from an associated Slack workspace.

    Outputs a JSON into OUT. Specify LOAD-FROM to merge with the output of a previous JSON to bypass
    the 10,000 message history.
    """
    try:
        conversations = list_conversations(course=course, include_dms=include_dms)
    except Exception as _:
        click.echo(
            "Unable to list conversations in Slack. "
            "Have you added your workspace to https://auth.cs61a.org and installed the bot at https://slack.cs61a.org?",
            err=True,
        )
        return

    print("Exporting Channels:", [c["name"] for c in conversations])
    click.confirm("Confirm export?", abort=True)

    export = {}

    if load_from:
        export = loads(load_from.read())

    for c in conversations:
        print(f"Exporting {c['name']}: ", end="", file=stderr, flush=True)
        messages = (
            {m["ts"]: m for m in export[c["id"]]["messages"]}
            if c["id"] in export
            else {}
        )
        new_messages = export_channel(course=course, channel_id=c["id"])
        print(len(new_messages), "messages downloaded", end="", file=stderr)
        for m in new_messages:
            messages[m["ts"]] = m
        print(",", len(messages), "total messages ", file=stderr)
        export[c["id"]] = dict(
            name=c["name"], messages=sorted(messages.values(), key=lambda m: m["ts"])
        )
    out.write(dumps(export))



from typing import List
from common.rpc.utils import create_service

service = create_service(__name__, "ag-worker")


@service.route("/api/ping_worker")
def ping_worker():
    ...


@service.route("/api/batch_grade")
def batch_grade(*, command: str, jobs: List[str], grading_zip: str):
    ...



from flask import current_app


def make_row(content, target, action="Remove"):
    """Create a form with value ``content`` that POSTs to ``target`` when the
    button labeled ``action`` is pressed.

    :param content: the body of the form
    :type content: str - HTML

    :param target: the URL to POST to
    :type target: str

    :param action: the label of the submit button
    :type action: str

    :return: a string representing the HTML form
    """
    return f"""<form style="display: inline" action="{target}" method="post">
            {content}
            <input type="submit" value="{action}">
    </form>"""


def html(out):
    """Adds some styling to the HTML body ``out``.

    Specifically, adds a header of the form "61A App Name" and prepends
    the SPCSS stylesheet (https://cdn.jsdelivr.net/npm/spcss@0.5.0).

    :param out: the original HTML
    :type out: str

    :return: a string representing the stylized HTML.
    """
    if "<h1>" not in out:
        if hasattr(current_app, "remote"):
            header = current_app.remote.consumer_key
            if header.startswith("61a-"):
                header = header[3:]
            header = header.replace("-", " ").title()
            out = f"<h1>61A {header}</h1>" + out
    return f"""
<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/spcss@0.5.0">
{out}
"""


def error(out):
    """Formats a page representing an error.

    Specifically, preformats the error message and adds a red header, with
    some instructions on who to contact for help.

    :param out: the error message
    :type out: str

    :return: a string representing the stylized HTML.
    """
    report = f"<pre>{out}</pre>" if out else ""
    return html(
        f"<h2 style='color: red'>Something went wrong.</h2>{report} "
        f'If you are on 61A staff, visit <a href="https://logs.cs61a.org">logs.cs61a.org</a> '
        f"to see crash logs. Otherwise, please post in the EECS Crossroads Slack for help."
    )



from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__, "code")


@requires_master_secret
@service.route("/api/create_code_shortlink")
def create_code_shortlink(
    name: str, contents: str, staff_only: bool = True, link: str = None
):
    ...



from os import getenv
from urllib.parse import urlparse

from flask import request, url_for as flask_url_for


def get_host() -> str:
    """Get the current host URL from request headers.

    :return: ``X-Forwarded-For-Host`` or ``Host``

    .. note::
        61A apps use ``X-Forwarded-For-Host``, which is not a standard header,
        for legacy reasons. This may be fixed in the future, but not for now.
    """
    return request.headers.get("X-Forwarded-For-Host") or request.headers.get("Host")


def url_for(*args, **kwargs) -> str:
    """Return the absolute URL target for the requested method.

    Uses :func:`~flask.url_for` to get the relative endpoint, then prepends the
    host from :func:`get_host` to the URL to make it absolute.

    All arguments are passed directly into :func:`~flask.url_for`.

    :return: the absolute target URL
    """
    host = get_host()
    redirect_url = urlparse(flask_url_for(*args, **kwargs))
    # noinspection PyProtectedMember
    return redirect_url._replace(
        netloc=host, scheme="https" if getenv("ENV") == "prod" else "http"
    ).geturl()
